/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.util;

import java.beans.PropertyVetoException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.openide.explorer.ExplorerManager;
import org.openide.nodes.Node;
import org.openide.util.Exceptions;

/**
 *
 * @author Timon Veenstra
 */
public class ExplorerManagerUtil {

    private static final Logger LOG = Logger.getLogger(ExplorerManagerUtil.class.getName());
    /**
     * add a bean to the node selection
     *
     * silently fails when object is not managed by the explorer manager
     *
     * @param object Bean to add to the selection
     */
    public static void addObjectToSelection(Object object, ExplorerManager explorerManager){
        try {
            List<Node> selected = new ArrayList<Node>(Arrays.asList(explorerManager.getSelectedNodes()));
            Node node = findNodeForObject(object, explorerManager.getRootContext());
            if (node == null) {
                LOG.log(Level.WARNING, "No node found for object {0} in explorer manager {1}", new Object[]{object, explorerManager});
                return;
            }
            // avoid duplicate selection
            if (!selected.contains(node)) {
                selected.add(node);
            }
            explorerManager.setSelectedNodes(selected.toArray(new Node[]{}));
        } catch (PropertyVetoException ex) {
            LOG.warning("addObjectToSelection vetoed by the explorer manager, a view might not be able to display the selected nodes");
        }
    }

    /**
     * remove a node from the selection
     * silently fails when object is not managed by the explorer manager
     *
     * @param object Bean to remove from the selection
     */
    public static void removeObjectFromSelection(Object object, ExplorerManager explorerManager) {
        try {
            List<Node> selected = new ArrayList<Node>(Arrays.asList(explorerManager.getSelectedNodes()));

            Node node = findNodeForObject(object, explorerManager.getRootContext());
            if (node == null) {
                LOG.log(Level.WARNING, "No node found for object {0} in explorer manager {1}", new Object[]{object, explorerManager});
                return;
            }

            if (selected.contains(node)) {
                selected.remove(node);
                explorerManager.setSelectedNodes(selected.toArray(new Node[]{}));
            }
        } catch (PropertyVetoException ex) {
            LOG.warning("addObjectToSelection vetoed by the explorer manager, a view might not be able to display the selected nodes");
        }
    }

    /**
     * Adds a node to the selection of an explorer manager.
     * When the node is already selected, nothing will happen.
     * silently fails when node is not managed by the explorer manager
     * 
     * @param node
     * @param explorerManager 
     */
    public static void addNodeToSelection(Node node, ExplorerManager explorerManager) {
        List<Node> selected = new ArrayList<Node>(Arrays.asList(explorerManager.getSelectedNodes()));
        if (!selected.contains(node)) {
            try {
                selected.add(node);
                explorerManager.setSelectedNodes(selected.toArray(new Node[]{}));
            } catch (PropertyVetoException ex) {
                LOG.warning("addObjectToSelection vetoed by the explorer manager, a view might not be able to display the selected nodes");
            }
        }
    }
    
    /**
     * removes a node from the selection of an explorer manager.
     * silently fails when node is not managed by the explorer manager
     * 
     * @param node
     * @param explorerManager 
     */
    public static void removeNodeFromSelection(Node node, ExplorerManager explorerManager) {
        List<Node> selected = new ArrayList<Node>(Arrays.asList(explorerManager.getSelectedNodes()));
        if (selected.contains(node)) {
            try {
                selected.remove(node);
                explorerManager.setSelectedNodes(selected.toArray(new Node[]{}));
            } catch (PropertyVetoException ex) {
                LOG.warning("addObjectToSelection vetoed by the explorer manager, a view might not be able to display the selected nodes");
            }
        }        
    }

    /**
     * Find the first node with the provided object in its lookup.
     * 
     * @param object 
     * @param node root node to start searching from
     * @return found node with object in its lookup
     */
    public static Node findNodeForObject(Object object, Node node) {
        if (object ==null || node == null){
            return null;
        }
        Object wrappedObject = node.getLookup().lookup(object.getClass());
        if (object.equals(wrappedObject)) {
            return node;
        }
        if (node.getChildren() != null) {
            for (Node child : node.getChildren().getNodes()) {
                Node found = findNodeForObject(object, child);
                if (found != null) {
                    return found;
                }
            }
        }
        return null;
    }
}
