/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.util.annotations.processor;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Processor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import nl.cloudfarming.client.util.annotationhandlers.AnnotationHandler;
import nl.cloudfarming.client.util.annotationhandlers.AnnotationUtils;
import nl.cloudfarming.client.util.annotations.ASGetter;
import nl.cloudfarming.client.util.annotations.ASSetter;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import nl.cloudfarming.client.util.FileUtil;
import nl.cloudfarming.client.util.annotationhandlers.ASEqualsHandler;
import nl.cloudfarming.client.util.annotationhandlers.ASGetterHandler;
import nl.cloudfarming.client.util.annotationhandlers.ASHashcodeHandler;
import nl.cloudfarming.client.util.annotationhandlers.ASSetterHandler;
import nl.cloudfarming.client.util.annotations.ASIdentifying;

/**
 *      http://netbeans.org/kb/docs/java/annotations-custom.html
 * 
 *      http://download.oracle.com/javase/1,5.0/docs/guide/apt/GettingStarted.html
 * 
 * http://tech-sash.blogspot.com/2010/04/gnerating-documentation-with-apt-and.html
 * http://download.oracle.com/javase/1.5.0/docs/guide/apt/mirror/index.html
 * 
 * 
 * http://openide.netbeans.org/tutorial/test-patterns.html
 * 
 * // FILTER?? : http://www.javalobby.org/java/forums/t17961.html (ONDERAAN)
 * 
 * Generate intermediate source file using the annotations.
 * 
 * @author Gerben Feenstra
 */
@SupportedAnnotationTypes("nl.cloudfarming.client.util.annotations.ASGetter")
@SupportedSourceVersion(SourceVersion.RELEASE_6)
@ServiceProvider(service = Processor.class)
public class ASAnnotationProcessor extends AbstractProcessor {

    List<AnnotatedClass> annotatedClasses = new ArrayList<AnnotatedClass>();

    /** public for ServiceLoader */
    public ASAnnotationProcessor() {
    }

    @Override
    public boolean process(Set annotations, RoundEnvironment roundEnv) {

        // Find all getters
        for (Element element : roundEnv.getElementsAnnotatedWith(ASGetter.class)) {
            TypeElement clazz = (TypeElement) element.getEnclosingElement();
            AnnotatedClass annotatedClass = getAnnotatedClazz(clazz);
            annotatedClass.addGetterField(element);
        }

        // Find all setters
        for (Element element : roundEnv.getElementsAnnotatedWith(ASSetter.class)) {
            TypeElement clazz = (TypeElement) element.getEnclosingElement();
            AnnotatedClass annotatedClass = getAnnotatedClazz(clazz);
            annotatedClass.addSetterField(element);
        }

        // Find all identifying fields
        for (Element element : roundEnv.getElementsAnnotatedWith(ASIdentifying.class)) {
            TypeElement clazz = (TypeElement) element.getEnclosingElement();
            AnnotatedClass annotatedClass = getAnnotatedClazz(clazz);
            annotatedClass.addIdentifyingField(element);
        }

        // Process AnnotatedClasses
        for (AnnotatedClass annotatedClass : annotatedClasses) {
            try {
                TypeElement clazz = annotatedClass.getClazz();

                String qualifiedSourceFileName = clazz.getQualifiedName().toString().replace("Dev", "");

                String dest = FileUtil.getPath(qualifiedSourceFileName) + ".java";
                File destFile = new File(dest);

                String sourceFileName = clazz.getSimpleName().toString().replace("Dev", "");

                FileOutputStream out = new FileOutputStream(destFile);

                // Write start of class
                try {
                    AnnotationUtils.writeLicense(out);
                    out.write(("package " + clazz.getEnclosingElement() + ";\n").getBytes());
                    out.write(("\n").getBytes());
                    out.write(("    /** GENERATED CLASS */\n").getBytes());
                    out.write(("public class " + sourceFileName + " extends " + clazz.getSimpleName() + " {\n").getBytes());
                    out.write(("\n").getBytes());
                    out.write(("    public " + sourceFileName + "(){}\n").getBytes());

                    // Write Getters
                    AnnotationHandler getterAnnotationHandler = getAnnotationHandler(ASGetterHandler.class);
                    for (Element getterAnnotatedElement : annotatedClass.getGetterFields()) {
                        getterAnnotationHandler.handle(out, getterAnnotatedElement);
                    }

                    // Write Setters
                    AnnotationHandler setterAnnotationHandler = getAnnotationHandler(ASSetterHandler.class);
                    for (Element setterAnnotatedElement : annotatedClass.getSetterFields()) {
                        setterAnnotationHandler.handle(out, setterAnnotatedElement);
                    }

                    // Identifying elements (Equals)
                    AnnotationHandler equalsAnnotationHandler = getAnnotationHandler(ASEqualsHandler.class);
                    try {
                        out.write(("@Override\n").getBytes());
                        out.write(("public boolean equals(Object obj) {\n").getBytes());
                        out.write(("    if(!super.equals(obj)){\n").getBytes());
                        out.write(("        return false;\n").getBytes());
                        out.write(("    }\n").getBytes());
                        out.write(("\n").getBytes());
                        out.write(("    if (obj == null) {\n").getBytes());
                        out.write(("        return false;\n").getBytes());
                        out.write(("    }\n").getBytes());
                        out.write(("\n").getBytes());
                        out.write(("    if (getClass() != obj.getClass()) {\n").getBytes());
                        out.write(("        return false;\n").getBytes());
                        out.write(("    }\n").getBytes());
                        out.write(("\n").getBytes());
                        out.write(("    final Field other = (Field) obj;\n").getBytes());
                        for (Element identifyingAnnotatedElement : annotatedClass.getIdentifyingFields()) {
                            equalsAnnotationHandler.handle(out, identifyingAnnotatedElement);
                        }
                        out.write(("\n").getBytes());
                        out.write(("    return true;\n").getBytes());
                        out.write(("}\n").getBytes());

                    } catch (IOException ex) {
                        Exceptions.printStackTrace(ex);
                    }

                    // Identifying elements (Hashcode)
                    AnnotationHandler hashCodeAnnotationHandler = getAnnotationHandler(ASHashcodeHandler.class);
                    try {
                        out.write(("@Override\n").getBytes());
                        out.write(("public int hashCode() {\n").getBytes());
                        out.write(("    int hash = super.hashCode();\n").getBytes());
                        for (Element identifyingAnnotatedElement : annotatedClass.getIdentifyingFields()) {
                            hashCodeAnnotationHandler.handle(out, identifyingAnnotatedElement);
                        }
                        out.write(("\n").getBytes());
                        out.write(("    return hash;\n").getBytes());
                        out.write(("}\n").getBytes());

                    } catch (IOException ex) {
                        Exceptions.printStackTrace(ex);
                    }
                } catch (IOException ex) {
                    Exceptions.printStackTrace(ex);
                }


            } catch (FileNotFoundException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
        return true;
    }

    /**
     * Find and return the annotationhandler instance or the specified annotationhandler
     * @param annotationHandlerToFind
     * @return 
     */
    private AnnotationHandler getAnnotationHandler(Class annotationHandlerToFind) {
        Collection<? extends AnnotationHandler> annotationHandlers = Lookup.getDefault().lookupAll(AnnotationHandler.class);
        for (AnnotationHandler annotationHandler : annotationHandlers) {
            if (annotationHandler.getClass().isAssignableFrom(annotationHandlerToFind)) {
                return annotationHandler;
            }
        }
        return null;
    }

    /**
     * Get the annotated class representing the specified TypeElement or create a new one in the List.
     * @param clazz
     * @return 
     */
    private AnnotatedClass getAnnotatedClazz(TypeElement clazz) {
        AnnotatedClass returnAnnotatedClass = null;
        for (AnnotatedClass annotatedClass : annotatedClasses) {
            if (annotatedClass.getClazz().equals(clazz)) {
                returnAnnotatedClass = annotatedClass;
                break;
            }
        }
        if (returnAnnotatedClass == null) {
            returnAnnotatedClass = new AnnotatedClass(clazz);
            annotatedClasses.add(returnAnnotatedClass);
        }
        return returnAnnotatedClass;
    }

    /**
     * An AnnotatedClass is a representation of which getters, setters and identifying field a class should have.
     */
    private class AnnotatedClass {

        private List<Element> getterFields = new ArrayList<Element>();
        private List<Element> setterFields = new ArrayList<Element>();
        private List<Element> identifyingFields = new ArrayList<Element>();
        private TypeElement clazz = null;

        public AnnotatedClass(TypeElement clazz) {
            this.clazz = clazz;
        }

        public void addGetterField(Element element) {
            getterFields.add(element);
        }

        public void addSetterField(Element element) {
            setterFields.add(element);
        }

        public void addIdentifyingField(Element element) {
            identifyingFields.add(element);
        }

        public TypeElement getClazz() {
            return this.clazz;
        }

        public List<Element> getGetterFields() {
            return this.getterFields;
        }

        public List<Element> getSetterFields() {
            return this.setterFields;
        }

        public List<Element> getIdentifyingFields() {
            return this.identifyingFields;
        }
    }
}
